import pandas as pd
from math import inf
import math
import sys
import shutil

n_days_of_trading_day = 250


def format_percent(p):
    return "%.2f%%" % (100 * p)


def format_ratio(r):
    return "%.2f" % r


def downside_dev(d):
    m = d.mean()
    s = 0
    for x in d:
        x = min(0, x - m)
        s += x * x
    n = len(d)
    s2 = s / (n - 1)
    return math.sqrt(s2)


class BackTesting:
    # ticker date detail
    daily_detail_panel = None
    iter_data = None

    stocks = None
    cash = 0

    commission_ratio = 2 / 1000

    money_series = None
    profit_series = None
    commission_series = None
    monthly_profit_series = None
    commission_per_share = False

    name = "test"

    # daily details
    amount_today = 0
    amount_yesterday = 0
    profit_today = 0
    close_prices_today = None
    open_prices_today = None
    i_days = 0
    current_date = None
    commission_today = 0

    # events control
    events = None  # what is done daily
    skip_morning = False
    skip_trade = False
    skip_afternoon = False
    skip_record = False
    skip_make = False
    skip_night = False

    # what to buy or what to sell
    buying = None
    selling = None
    throwing = None  # which will be elimited on

    # for a progress bar
    total_days = 0
    display_progress = True
    bar_width = 0

    def __init__(self, name='test', cash=0, daily_detail_panel=pd.Panel(), **kwargs):
        self.bar_width = shutil.get_terminal_size()[0]
        self.daily_detail_panel = daily_detail_panel
        self.name = name
        self.cash = cash
        self.strategy = pd.Series()

        self.stocks = pd.Series()
        self.money_series = pd.Series()
        self.profit_series = pd.Series()
        self.commission_series = pd.Series()
        self.monthly_profit_series = pd.Series()

        self.buying = pd.Series()
        self.selling = pd.Series()
        self.throwing = pd.Series()

        # add extra attributes
        for k, v in kwargs.items():
            setattr(self, k, v)

        self.events = [
            (self.morning, 'skip_morning'),
            (self.trade, 'skip_trade'),
            (self.afternoon, 'skip_afternoon'),
            (self.record, 'skip_record'),
            (self.make, 'skip_make'),
            (self.night, 'skip_night')
        ]

    def init(self):
        pass

    def prune(self):
        # ticker date detail
        # we need date ticker detail
        # self.iter_data = self.daily_detail_panel.transpose(1, 0, 2)
        self.iter_data = self.daily_detail_panel
        self.total_days = len(self.iter_data)

    def rows(self):
        # data = data.rename(items={'open': 'open_prices'})
        return self.iter_data.iteritems()

    def parse_line(self, line):
        (d, v) = line
        self.current_date = d
        self.open_prices_today = v['open']
        self.close_prices_today = v['close']

    def run(self):
        self.prune()
        self.init()
        if self.display_progress:
            print('progress')
            sys.stdout.write("[%s]" % (" " * (self.bar_width - 2)))
            sys.stdout.write("\b" * (self.bar_width - 2))  # return to start of line, after '['
            sys.stdout.write("  0.0%")
            sys.stdout.flush()
        d = int(self.total_days / (self.bar_width - 2))

        for i, line in enumerate(self.rows()):
            # show the progress
            if self.display_progress:
                if i % d == 0:
                    sys.stdout.write('\b\b\b\b\b\b# %04.1f%%' % (i / self.total_days * 100))
                    sys.stdout.flush()

            self.i_days = i
            self.parse_line(line)
            for (e, s) in self.events:
                if getattr(self, s):
                    setattr(self, s, False)
                    continue
                if not e():  # not success
                    break

        if self.display_progress:
            sys.stdout.write('\b\b\b\b\b\b# 100%\n')
            sys.stdout.flush()

    # events
    def morning(self):
        return True

    def trade(self):
        commission = 0

        for ticker, number in self.selling.iteritems():
            price = self.open_prices_today[ticker]
            if self.commission_per_share:
                r = self.commission_ratio / price
            else:
                r = self.commission_ratio

            self.stocks[ticker] -= number
            commission += price * number * r
            self.cash += price * number * (1-r)  # subtract the commission

        for ticker in self.throwing.iteritems():
            price = self.open_prices_today[ticker]
            number = self.stocks[ticker]
            if self.commission_per_share:
                r = self.commission_ratio / price
            else:
                r = self.commission_ratio

            del self.stocks[ticker]  # no more needed
            commission += price * number * r
            self.cash += price * number * (1-r)  # subtract the commission

        for ticker, money in self.buying.iteritems():
            price = self.open_prices_today[ticker]
            if self.commission_per_share:
                r = self.commission_ratio / price
            else:
                r = self.commission_ratio

            self.stocks[ticker] += money / price * (1 - r)

            self.cash -= money
            commission += r * money

        commission += self.adjust()

        self.commission_today = commission

        # clean the strategies since they are no longer needed
        self.buying = pd.Series()
        self.selling = pd.Series()
        self.throwing = pd.Series()
        self.strategy = pd.Series()
        return True

    def afternoon(self):
        # calculate daily info
        self.calculate_daily_info()
        return True

    def record(self):
        d = self.current_date
        self.money_series[d] = self.amount_today
        self.commission_series[d] = self.commission_today
        self.profit_series[d] = self.profit_today
        return True

    def night(self):
        self.amount_yesterday = self.amount_today
        self.commission_today = 0

        return True

    def make(self):
        # do nothing default
        self.buying = pd.Series()
        self.selling = pd.Series()
        self.strategy = pd.Series()
        return True

    # daily utils
    def calculate_daily_info(self):
        stock_value = 0
        for ticker, number in self.stocks.iteritems():
            price = self.close_prices_today[ticker]
            stock_value += number * price
        self.amount_today = self.cash + stock_value
        self.profit_today = (self.amount_today - self.amount_yesterday) / self.amount_yesterday

    def adjust(self):
        if len(self.strategy) == 0:  # we do not need to process
            return 0
        # Before trading, we get some cash, n etfs(number of share denoted by ei),
        # and the open price for each eft(denoted by oi).
        # Also, we know the number of shares per etf,
        # so we can get an amount at first.
        # amount = sum([e * p for e, p in zip(self.stocks, self.open_prices_today)]) + self.cash
        amount = self.cash
        for ticker, number in self.stocks.iteritems():
            price = self.open_prices_today[ticker]
            amount += number * price

        # Let v be the total amount after trading.
        # Then we should decide how many shares we need to buy or throw per etf
        # (positive for buying and negative for throwing, namely the delta).
        # denoted by d1, d2, ..., dn.
        # During trading, we need to pay for commission,
        # so after that, we will find the following equation
        # that amount = r1 * |d1| + r2 * |d2| + ... + rn * |dn| + v,
        # where r is the commission ratio per share and
        # |x| means the absolute value of x.
        # Here we need absolute value because no matter we buy or throw a etf,
        # the commission is charged per share, and hence the total value is reduced
        # we can say that the efts whose value exceeds it's target percentage need
        # to be reduced, but cannot decide others should be increased or not no reduced,
        # so we could only remove part of those absolute values
        # Finally, after trading, we should get the following result that
        # (ei + di) * oi = pi * v, i = 1, 2, ..., n
        # To solve the equation, we've got (n+1) unknowns and (n+1) equations,
        # n of which are linear ones, so we can use v to represent di_s to reduce
        # the number of unknown to 1, and we get a linear combination of v and ai|v - bi|,
        # where ai and bi can be decided with some transformation and the equation
        # can be solved in (n+1) times which reduces the time complexity

        # amount = r1 * |d1| + r2 * |d2| + ... + rn * |dn| + v
        # di = pi * v / oi - ei
        # if pi == 0, di = -ei
        # and then we will add r * ei to the right of the equation
        # if pi != 0 then we can get
        # di = pi / oi * (v - ei * oi / pi)
        # let ai = pi / oi * ri, bi = ei * oi / pi
        # then ri * di = ai * (v - bi)
        # amount = k * v + const + sigma ai * |v - bi|
        k = 1
        c = 0
        # since the order of ai bi does not affect the amount
        ba_list = []
        for ticker, p in self.strategy.iteritems():
            e = self.stocks[ticker]
            o = self.open_prices_today[ticker]
            if not self.commission_per_share:  # I need r which is commission per share
                r = self.commission_ratio * o
            else:
                r = self.commission_ratio
            if p == 0:  # |di| = ei
                c += r * e
                continue
            a = p / o * r
            b = e * o / p
            if e * o > p * amount:  # we need to reduce the percentage, namely r * |di| = - r * di = ai(bi - v)
                k += -a
                c += a * b
            else:
                # we need absolute value
                ba_list.append((b, -a))  # ai are negative first in the following calculation
                # here I use a list but not a tuple because list is mutable
                # I need to change the value of a each time traverse an interval
        c = amount - c
        # c = k * v + sigma ai * |v - bi|
        if len(ba_list) == 0:  # no more extra calculation
            v = c / k
        else:
            ba_list.sort()  # since the first item is the key
            a_list = [[ba[1]] for ba in ba_list]  # I need a mutable object
            b_list = [ba[0] for ba in ba_list]

            v = 0  # since the amount must be positive
            # -inf, b1
            # b1, b2
            # b2, b3
            # ...
            # bn-1, bn
            # bn, inf
            b1_list = [-inf] + b_list
            b2_list = b_list + [inf]
            a_list.append([0])  # the b1,b2 list loops n + 1 times, we need avoid error
            b_list.append(0)
            ia = iter(a_list)
            for b1, b2 in zip(b1_list, b2_list):
                # c = k * v + sigma ai * (v - bi) = (k + sigma a) * v - sigma ai * bi
                sa = sum([a for [a] in a_list])
                samb = sum([a*b for [a], b in zip(a_list, b_list)])
                d = k + sa
                s = c + samb
                # d * v = s
                # v = s / d
                new_v = v
                if d == 0:
                    if s == 0:
                        new_v = 0
                else:
                    t = s / d
                    if b1 - 0.000001 <= t <= b2 + 0.000001:
                        new_v = s / d

                v = max(v, new_v)
                # change the value of the current value
                a = next(ia)
                a[0] = -a[0]

        for ticker, p in self.strategy.iteritems():
            o = self.open_prices_today[ticker]
            self.stocks[ticker] = v * p / o
        self.cash = v * (1 - sum(self.strategy))
        # commission during this progress is amount - v
        return amount - v  # return the commission for further calculate

    # other helping functions
    def get_monthly_profit(self):
        curr_month = self.money_series.index[0].month
        s = self.money_series.iloc[0]
        e = self.money_series.iloc[0]
        ld = None
        for d, v in self.money_series.iteritems():
            m = d.month
            ld = d
            if m == curr_month:  # in the same month
                e = v  # newer the end of the month
            else:  # change into another month
                # we need to update max return
                r = (e - s) / s
                s = e = v
                curr_month = m
                self.monthly_profit_series[d] = r
        # the last month is not included, so we need to do once more
        r = (e - s) / s
        self.monthly_profit_series[ld] = r

    # calculate all statistics
    def statistics(self, rf=0):
        self.get_monthly_profit()
        result = pd.Series()
        result['CAGR'] = format_percent(self.cagr())
        result['STD DEV'] = format_percent(self.std_dev())
        result['DOWNSIDE DEV'] = format_percent(self.downside_dev())
        result['SHARPE RATIO'] = format_ratio(self.sharpe_ratio(rf))
        result['SORTINO RATIO'] = format_ratio(self.sortino_ratio(rf))
        result['WORST DRAWDOWN'] = format_percent(self.worst_drawdown())
        result['WORST MONTH RETURN'] = format_percent(self.monthly_profit_series.min())
        result['BEST MONTH RETURN'] = format_percent(self.monthly_profit_series.max())
        result['%PROFITABLE MONTH'] = format_percent(self.profitable_month())
        result['COMMISSION'] = sum(self.commission_series)
        result['START DATE'] = self.money_series.index[0].date()
        result['END DATE'] = self.money_series.index[-1].date()
        return result

    # I need to calculate the following statistics
    def cagr(self, units_of_a_year=n_days_of_trading_day):
        # units_of_a_year means what's contained in the data_list
        # with 250 for day and 12 for month
        data = self.money_series
        origin = data.iloc[0]
        now = data.iloc[-1]
        number = 0
        min_date = data.index[0]
        max_date = data.index[-1]
        min_year_ndays = 0  # number of days of the minimum year
        max_year_ndays = 0  # number of days of the maximum year
        curr_year = min_date.year
        for d in data.index:
            y = d.year
            if y == min_date.year:
                min_year_ndays += 1
            elif y == max_date.year:
                max_year_ndays += 1
            else:
                if curr_year != y:  # indicate a new year
                    number += 1
                    curr_year = y  # update current year

        if min_date.year == max_date.year:  # means there is only a single year
            number = min_year_ndays / units_of_a_year
        # I don't put follow condition in the for loop before in order to speed up :)
        elif max_date.year - min_date.year == 1:  # only two years so that there's no need to add
            number = (min_year_ndays + max_year_ndays) / units_of_a_year
        else:
            number += (min_year_ndays + max_year_ndays) / units_of_a_year
        if number == 0:
            print("Too few data to calculate gagr")
            exit(-1)
        return (now / origin) ** (1 / number) - 1

    def std_dev(self):
        return self.monthly_profit_series.std()

    def downside_dev(self):
        return downside_dev(self.monthly_profit_series)

    def sharpe_ratio(self, rf):
        s = self.monthly_profit_series.std()
        e = self.monthly_profit_series.mean()
        return (e - rf) / s * math.sqrt(12)

    def sortino_ratio(self, rf):
        s = downside_dev(self.monthly_profit_series)
        e = self.monthly_profit_series.mean()
        return (e - rf) / s * math.sqrt(12)

    def worst_drawdown(self):
        data = self.money_series
        min_after = data.iloc[-1]
        drawdown = 0

        # I don't know whether to use .iloc[::-1] or just in this way
        # thinking that reversed may use a bi-direction list so that it
        # will be faster
        for v in reversed(data):
            min_after = min(min_after, v)
            # calculate a percentage
            drawdown = max(drawdown, (v - min_after) / v)
        return drawdown

    def profitable_month(self):
        def trigger(x):
            if x > 0:
                return 1
            else:
                return 0
        profitable_months = self.monthly_profit_series.map(trigger)
        return sum(profitable_months) / len(profitable_months)
