#!/usr/bin/env python3
import os
import pandas as pd
import matplotlib.pyplot as plt


from backtesting import BackTesting

import strategies


def main():
    data_dir = 'data'
    output_dir = 'output'
    ticker = '399001'
    plt.close()
    img = pd.DataFrame()
    df = pd.read_pickle( ticker + '.pkl')
    panel = pd.Panel(data={ticker: df}).transpose(1, 0, 2)

    n_days_list = [5, 10, 20, 60, 120, 250]

    m_n_days = max(n_days_list)
    plain = strategies.PlainBuying(start_iloc=m_n_days, daily_detail_panel=panel, cash=1000000)
    plain.stocks = pd.Series(index=[ticker], data=[0])
    print('BUY AND HOLD')
    plain.run()
    print(plain.statistics())

    statistics = pd.DataFrame()
    statistics[ticker + '_BUY AND HOLD'] = plain.statistics()
    img['BUY AND HOLD'] = plain.money_series

    def get_strategy(n_days):
        new_s = strategies.Robust(
            start_iloc=m_n_days,
            name='s'+str(n_days), n_days=n_days,
            indicator=ticker,
            daily_detail_panel=panel,
            cash=1000000
        )
        new_s.stocks = pd.Series(index=[ticker], data=[0])
        return new_s

    for n in n_days_list:
        s = get_strategy(n)
        print(s.name)
        s.run()
        img[s.name] = s.money_series
        print(s.statistics())
        statistics[ticker + '_' + s.name] = s.statistics()
    img.plot()
    plt.savefig(output_dir + os.sep + ticker + '.pdf')
    statistics.to_excel(output_dir + os.sep + 'statistics.xlsx')


if __name__ == '__main__':
    main()
