import pandas as pd

from backtesting import BackTesting


class Robust(BackTesting):
    start_iloc = 0
    need_balance = False
    balanced_days = None
    balance_period = 1
    n_days = 1

    indicator_close_prices = None
    icp = 0
    moving_averages = None
    ma = 0
    time_series_momenta = None
    tsmon = 0

    last_strategy = 0
    indicator = "000001"

    def __init__(self, n_days=1, start_iloc=0, indicator="000001", *args, **kwargs):
        super(Robust, self).__init__(*args, **kwargs)
        self.commission_ratio = 2 / 1000
        self.start_iloc = start_iloc
        self.n_days = n_days
        close_series = self.daily_detail_panel.loc[:, indicator, 'close']
        self.indicator_close_prices = close_series
        r = close_series.rolling(n_days)
        self.moving_averages = r.mean()
        prev_close = r.apply(lambda x: x[0])
        self.time_series_momenta = (close_series - prev_close) / close_series
        self.indicator = indicator

    def init(self):
        super(Robust, self).init()
        self.amount_yesterday = self.cash
        self.skip_trade = True  # we need to do nothing from the day to make a decision

    def prune(self):
        super(Robust, self).prune()

        self.time_series_momenta = self.time_series_momenta[self.start_iloc - 1:]
        self.moving_averages = self.moving_averages[self.start_iloc - 1:]
        self.indicator_close_prices = self.indicator_close_prices[self.start_iloc - 1:]
        self.iter_data = self.iter_data.iloc[self.start_iloc - 1:]

    def rows(self):
        return zip(
            self.moving_averages,
            self.time_series_momenta,
            self.indicator_close_prices,
            self.iter_data.iteritems()
        )

    def parse_line(self, line):
        ma, tsmon, icp, old_line = line
        self.ma = ma
        self.tsmon = tsmon
        self.icp = icp
        super(Robust, self).parse_line(old_line)

    def make(self):
        n = len(self.stocks)
        m = self.cash / n
        ma = self.ma
        tsmon = self.tsmon
        price = self.icp

        # if price > ma and tsmon > 0:  # 1
        #     self.buying = pd.Series(index=self.stocks.index, data=m)  # buy them all
        #     self.last_strategy = 1
        # elif price > ma or tsmon > 0:
        #     if self.last_strategy == 1:  # we need to sell
        #         self.selling = self.stocks / 2  # with a half left
        #     elif self.last_strategy == 0:
        #         self.buying = pd.Series(index=self.stocks.index, data=m/2)  # buy with half of the money
        #     else:  # the same as yesterday
        #         pass  # do nothing
        #     self.last_strategy = 0.5
        # else:
        #     self.selling = self.stocks  # sell them all
        #     self.last_strategy = 0
        if price > ma:
            self.buying = pd.Series(index=self.stocks.index, data=m)
        else:
            self.selling = self.stocks

        return True

    def run(self):
        super(Robust, self).run()
        self.money_series = self.money_series.iloc[1:]
        self.profit_series = self.profit_series.iloc[1:]
        self.commission_series = self.commission_series[1:]
